#include <stdio.h>

#include "juggle.h"

#define extern
#include "siteswap.h"
#undef extern

#include "globals.h"

boolean 
mark_catch(long array[], long position)
/*{{{  */
{
    long mask, offset;

    mask = 1 << (position % BITS_PER_LONG);
    offset = position / BITS_PER_LONG;
    if ((array[offset] & mask) != 0)
	return(0);		/* collision */
    array[offset] |= mask;
    return(1);
}
/*}}}  */

void 
check_pattern(siteswap *ss)
/*{{{  */
{
   long pos, seq, held, limit, throw;
   int  s = ss->startup_len;
   long used[TOTAL_PATTERN_LENGTH/BITS_PER_LONG + 1] = {0,};
   char *where;
   char msg[256];
   
   held = ss->balls;
   limit = s + max(ss->pattern_len, ss->highest);
   for (pos = 0; pos < limit; pos++) 
   {
      if (held > 0) 
      {
         if (!mark_catch(used, pos))
         {
            sprintf(msg,
               "Oops, a ball lands while %d ball%s not yet airborne\n",
               held,
               held == 1 ? " is" : "s are");
            show_msg(msg, TRUE);
            return;
         }
      }  
      
      if (pos < s)
         throw = ss->pattern[MAX_PATTERN_LENGTH - s + pos];
      else
         throw = ss->pattern[MAX_PATTERN_LENGTH + ((pos - s) % ss->pattern_len)];
      
      if (throw != 0 || held <= 0) 
      {	/* "empty" during launch means just hold */
         held--;		/* a ball was really launched here */
         if (!mark_catch(used, throw + pos))
         {
            /*closegraph();*/
            if (pos < s) 
            {
               where = "startup";
               seq = pos + 1;
            } 
            else 
            {
               where = "pattern";
               seq = pos + 1 - s;
            } 
  
            if (throw == 0)
               sprintf(msg,
                  "Oops, an earlier throw hits an empty hand at %s throw %d\n",
                  where, seq);
            else
               sprintf(msg,
                  "Oops, %s throw %d (a %d) hits a hand holding an earlier throw\n",
                  where, seq, throw);
            show_msg(msg,TRUE);
            return;
         }
      }	
   }  
      
   for (pos = 0; pos < ss->balls; pos++) 
   {
      if (mark_catch(used, pos + s)) 
      {
         sprintf(msg,
            "Oops, no throw lands in position %d after first cycle\n",
            pos + 1);
         show_msg(msg,TRUE);
         return;
      }	   
   }
   
   ss->valid = TRUE;
}   
/*}}}  */

int 
find_init_pattern(siteswap *ss)
/*{{{  */
{
   long pos, seq, limit, throw;
   int  s = ss->startup_len;
   long used[TOTAL_PATTERN_LENGTH/BITS_PER_LONG + 1] = {0,};
   char msg[256];
   
   limit = ss->pattern_len + ss->highest;
   
   for (pos = 0; pos < limit; pos++) 
   {
      throw = ss->pattern[MAX_PATTERN_LENGTH + (pos % ss->pattern_len)];
      if (!mark_catch(used, throw + pos)) 
      {
         seq = pos + 1;
         if (throw == 0)
            sprintf(msg,
			"Oops, an earlier throw hits an empty hand at pattern throw %d\n",
			seq);
         else
            sprintf(msg,
			"Oops, pattern throw %d (a %d) hits a hand holding an earlier throw\n",
			seq, throw);
	 show_msg(msg,TRUE);
         return 0;
      }
   }
   
   for (pos = limit - 1; pos >= 0 && pos >= ss->balls - s; pos--)
   {
      if (mark_catch(used, pos)) 
      {
         s++;
         ss->pattern[MAX_PATTERN_LENGTH - s] = pos + s;
      }
   }

   if (s > 0) 
   {
      sprintf(msg,"Generated startup: ");
      for (pos = 0; pos < s; pos++)
	    sprintf(msg+strlen(msg)," %d", ss->pattern[MAX_PATTERN_LENGTH + pos - s]);
      strcat(msg,"\n");
      show_msg(msg,FALSE);
   }
   
   ss->startup_len = s;
}
/*}}}  */

int 
string_to_pat(char *s, long pat[])
/*{{{  */
{
   int val,
       i      = 0,
       throws = 0,
       len    = strlen(s);
   
   while(i < len)
   {
      val = 0;
   
      while( i<len && isspace(s[i]) ) i++;   /* Skip leading space */
   
      if ( i<len && !isdigit(s[i]) )
      {
         show_msg("Garbage in pattern!\n",TRUE);
         return;
      }
      
      if (i<len)
      {
         while( i<len && isdigit(s[i]) )
         {
            val = 10*val + s[i] - '0';
            i++;
         }
         
         pat[throws++] = val;
      }
   }
   
   return throws;
}
/*}}}  */

void 
extract_pattern(siteswap* ss)
/*{{{  */
{
   int    i, sum;
   char   msg[256];
   
   ss->pattern_len = string_to_pat(ss->pattern_str, 
                                   &ss->pattern[MAX_PATTERN_LENGTH]);
   
   if (ss->pattern_len == 0)
      return;
      
   if (ss->startup_str[0] == '\0')
      ss->startup_len = 0;
   else
      string_to_pat(ss->startup_str, 
                    &ss->pattern[MAX_PATTERN_LENGTH - ss->startup_len]);

   ss->highest = 0; sum = 0;
   for(i=0; i<ss->pattern_len; i++)
   {
      sum         += ss->pattern[MAX_PATTERN_LENGTH + i];
      ss->highest =  max(ss->highest, ss->pattern[MAX_PATTERN_LENGTH + i]);
   }

   ss->balls = sum/ss->pattern_len;
   if ( sum != ss->balls*ss->pattern_len )
   {
      sprintf(msg,"Total %d not divisible by %d\n", sum, ss->pattern_len);
      show_msg(msg,TRUE);
      return;
   }

   if (ss->startup_len == 0)
      find_init_pattern(ss);
    
   ss->prehighest = ss->highest;
   for(i=0; i<ss->startup_len; i++)
      ss->prehighest = max(ss->prehighest, 
                           ss->pattern[MAX_PATTERN_LENGTH-i]);
    
   check_pattern(ss);

   ss->detail = 0; 
   ss->height = 40;
   
   return;
}
/*}}}  */

